//-----------------------------------------------------------------------------
//  Mandelbrot Set Generator
//  ECE573 Project, Winter 2008
//  Author:  Jesse Armagost
//-----------------------------------------------------------------------------

`timescale 1 ps / 1 ps

`include "mand_defs.v"

`define PIPE_STAGES 9

module complex_pipe ( input                 reset_n,
                      input                 pipeclk,
                      input [`DIGITS - 1:0] c_re_in,
                      input [`DIGITS - 1:0] c_im_in,
                      input [9:0]           v_pixel_in,
                      input [9:0]           h_pixel_in,
                      input                 load_en,
                      output                busy,
                      output                full,
                      output                done,
                      output [15:0]         iterations,
                      output                in_set,
                      output [9:0]          v_pixel_out,
                      output [9:0]          h_pixel_out   );


  //regs and wires
  reg [`DIGITS - 1:0] c_re_q [`PIPE_STAGES - 1:0];
  reg [`DIGITS - 1:0] c_im_q [`PIPE_STAGES - 1:0];
  reg [9:0] v_pixel_q [`PIPE_STAGES - 1:0];
  reg [9:0] h_pixel_q [`PIPE_STAGES - 1:0];
  reg [`PIPE_STAGES - 1:0] valid_q;
  reg [15:0] iter_count_q [`PIPE_STAGES - 1:0];
  integer i;

  //these signals are for the (z = z^2 + c) and magnitude squared logic.

  //stage 1
  reg [`DIGITS - 1:0] z_re;
  reg [`DIGITS - 1:0] z_im;
  //stage 5
  wire [(`DIGITS * 2) - 1:0] mult_3;
  wire [(`DIGITS * 2) - 1:0] mult_3_norm;
  wire [(`DIGITS * 2) - 1:0] mult_1;
  wire [(`DIGITS * 2) - 1:0] mult_1_norm;
  wire [(`DIGITS * 2) - 1:0] mult_2;
  wire [(`DIGITS * 2) - 1:0] mult_2_norm;
  wire [`DIGITS - 1:0] c_im_5q;
  wire [(`DIGITS * 2) - 1:0] c_im_5q_se;
  //stage 7
  wire [(`DIGITS * 2) - 1:0] add_2;
  wire [`DIGITS - 1:0] add_2_trunc;
  wire [(`DIGITS * 2) - 1:0] sub_1;
  wire [(`DIGITS * 2) - 1:0] add_3;
  wire [(`DIGITS * 1) - 1:0] c_re_7q;
  wire [(`DIGITS * 2) - 1:0] c_re_7q_se;
  wire add_2_of;
  wire sub_1_of;
  wire add_3_of;
  //stage 8
  reg [`DIGITS - 1:0] add_2_trunc_1q;
  reg of_a;
  reg mag_gt_four;
  //stage 9
  reg [`DIGITS - 1:0] result_im;
  wire [(`DIGITS * 2) - 1:0] add_1;
  wire [`DIGITS - 1:0] result_re;
  wire add_1_of;
  reg of_b;
  wire overflow;
  reg iter_done;



  //-----------------------
  //pipeline overhead logic
  //-----------------------



  //create the loadable complex plane coordinate registers.  there is one pair
  //  of real and imaginary coordinates per pipe stage.  when load_en goes
  //  high we load in the new re and im coordinates, else they are piped along.
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n) begin
      for (i = 0; i < `PIPE_STAGES; i = i + 1) begin
        c_re_q[i] <= 0;
        c_im_q[i] <= 0;
      end
    end
    else if (pipeclk) begin
      if (load_en) begin
        c_re_q[0] <= c_re_in;
        c_im_q[0] <= c_im_in;
      end
      else begin
        c_re_q[0] <= c_re_q[`PIPE_STAGES - 1];
        c_im_q[0] <= c_im_q[`PIPE_STAGES - 1];
      end
      for (i = 1; i < `PIPE_STAGES; i = i + 1) begin
        c_re_q[i] <= c_re_q[i - 1];
        c_im_q[i] <= c_im_q[i - 1];
      end
    end
  end

  //create the loadable pixel registers.  these registers just keep track of
  //  h_pixel and v_pixel values that correspond to the complex coordinate c.
  //  when load_en is high we load the new pixel values into the front of the
  //  pipe.
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n) begin
      for (i = 0; i < `PIPE_STAGES; i = i + 1) begin
        v_pixel_q[i] <= 0;
        h_pixel_q[i] <= 0;
      end
    end
    else if (pipeclk) begin
      if (load_en) begin
        v_pixel_q[0] <= v_pixel_in;
        h_pixel_q[0] <= h_pixel_in;
      end
      else begin
        v_pixel_q[0] <= v_pixel_q[`PIPE_STAGES - 1];
        h_pixel_q[0] <= h_pixel_q[`PIPE_STAGES - 1];
      end
      for (i = 1; i < `PIPE_STAGES; i = i + 1) begin
        v_pixel_q[i] <= v_pixel_q[i - 1];
        h_pixel_q[i] <= h_pixel_q[i - 1];
      end
    end
  end

  assign v_pixel_out = v_pixel_q[`PIPE_STAGES - 1];
  assign h_pixel_out = h_pixel_q[`PIPE_STAGES - 1];

  //create the valid bits pipe.  once a complex coordinate is done in the pipe,
  //  if the load signal is not active to load in a new value, the valid bit is
  //  cleared so we don't perform another erroneous write again because of the
  //  junk left in the pipeline.
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n) begin
      for (i = 0; i < `PIPE_STAGES; i = i + 1)
        valid_q[i] <= 0;
    end
    else if (pipeclk) begin
      if (load_en) begin
        valid_q[0] <= 1'b1;
      end
      else if (done) begin
        valid_q[0] <= 1'b0;
      end
      else begin
        valid_q[0] <= valid_q[`PIPE_STAGES - 1];
      end
      for (i = 1; i < `PIPE_STAGES; i = i + 1)
        valid_q[i] <= valid_q[i - 1];
    end
  end

  //the complex pipe is busy if the stage at the end of the pipe is currently
  //  valid.
  assign busy = valid_q[`PIPE_STAGES - 1];

  //indicate when the pipe is full, with no bubbles
  assign full = &valid_q;

  //create the clearable iteration count registers.  if we are loading a new
  //  value into the front of the pipe, we clear the iteration count.
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n) begin
      for (i = 0; i < `PIPE_STAGES; i = i + 1)
        iter_count_q[i] <= 0;
    end
    else if (pipeclk) begin
      if (load_en) begin
        iter_count_q[0] <= 0;
      end
      else begin
        iter_count_q[0] <= iter_count_q[`PIPE_STAGES - 1] + 1'b1;
      end
      for (i = 1; i < `PIPE_STAGES; i = i + 1)
        iter_count_q[i] <= iter_count_q[i - 1];
    end
  end

  assign iterations = iter_count_q[`PIPE_STAGES - 1];

  //provide the loading ability for z at the front of the pipe.  if we're not
  //  loading then just loop around the value from the end of the pipe.
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n) begin
      z_re <= 0;
      z_im <= 0;
    end
    else if (pipeclk) begin
      if (load_en) begin
        z_re <= c_re_in;
        z_im <= c_im_in;
      end
      else begin
        z_re <= result_re;
        z_im <= result_im;
      end
    end
  end



  //-----------------------
  //complex iteration logic
  //-----------------------



  //register the truncated result of add_2 twice to satisfy pipeline timing
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n) begin
      add_2_trunc_1q <= 0;
      result_im <= 0;
    end
    else if (pipeclk) begin
      add_2_trunc_1q <= add_2_trunc;
      result_im <= add_2_trunc_1q;
    end
  end

  //register the combined overflow signal composed of overflows from add_2,
  //  sub_1 and add_3.
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n)
      of_a <= 0;
    else if (pipeclk)
      of_a <= (  add_2_of
               | sub_1_of
               | add_3_of  );
  end

  //register the magnitude greater than 4 signal
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n)
      mag_gt_four <= 0;
    else if (pipeclk) begin
      if (add_3 > (1 << (`FRAC_DIGITS + 2)))
        mag_gt_four <= 1'b1;
      else
        mag_gt_four <= 1'b0;
    end
  end

  //register the of_b signal
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n)
      of_b <= 1'b0;
    else if (pipeclk)
      of_b <= (of_a | mag_gt_four );
  end

  //register the iter_done signal
  always @(negedge reset_n or posedge pipeclk) begin
    if (~reset_n)
      iter_done <= 1'b0;
    else if (pipeclk) begin
      if (iter_count_q[7] == `MAX_ITER)
        iter_done <= 1'b1;
      else
        iter_done <= 1'b0;
    end
  end

  //create a 32 bit sign-extended version of c_re
  assign c_re_7q = c_re_q[6];
  assign c_re_7q_se = (c_re_7q[`DIGITS - 1])
                    ? {{`DIGITS{1'b1}}, c_re_7q}
                    : {{`DIGITS{1'b0}}, c_re_7q};

  //create a 32 bit sign-extended version of c_im
  assign c_im_5q = c_im_q[4];
  assign c_im_5q_se = (c_im_5q[`DIGITS - 1])
                    ? {{`DIGITS{1'b1}}, c_im_5q}
                    : {{`DIGITS{1'b0}}, c_im_5q};

  //instantiate a mult module for (z_re ^ 2)
  mult i_mult_1 ( .clock (pipeclk),
                  .dataa (z_re),
                  .datab (z_re),
                  .result (mult_1)  );

  //normalize the mult_1 result with sign extension
  assign mult_1_norm = (mult_1[2 * `DIGITS - 1])
                ? {{`FRAC_DIGITS{1'b1}}, mult_1[2 * `DIGITS - 1:`FRAC_DIGITS]}
                : {{`FRAC_DIGITS{1'b0}}, mult_1[2 * `DIGITS - 1:`FRAC_DIGITS]};

  //instantiate a mult module for (z_im ^ 2)
  mult i_mult_2 ( .clock (pipeclk),
                  .dataa (z_im),
                  .datab (z_im),
                  .result (mult_2)  );

  //normalize the mult_2 result with sign extension
  assign mult_2_norm = (mult_1[2 * `DIGITS - 1])
                ? {{`FRAC_DIGITS{1'b1}}, mult_2[2 * `DIGITS - 1:`FRAC_DIGITS]}
                : {{`FRAC_DIGITS{1'b0}}, mult_2[2 * `DIGITS - 1:`FRAC_DIGITS]};

  //instantiate a mult module for z_re * z_im
  mult i_mult_3 ( .clock (pipeclk),
                  .dataa (z_re),
                  .datab (z_im),
                  .result (mult_3)  );

  //normalize the mult_3 result with sign extension.  we need to multiply by 2
  //  anyway, so we shift right one less than usual.
  assign mult_3_norm = (mult_3[2 * `DIGITS - 1])
        ? {{`FRAC_DIGITS - 1{1'b1}}, mult_3[2 * `DIGITS - 1:`FRAC_DIGITS - 1]}
        : {{`FRAC_DIGITS - 1{1'b0}}, mult_3[2 * `DIGITS - 1:`FRAC_DIGITS - 1]};

  //instantiate a subtractor to create z_re^2 - z_im^2
  sub i_sub_1 ( .clock (pipeclk),
                .dataa (mult_1_norm),
                .datab (mult_2_norm),
                .overflow (sub_1_of),
                .result (sub_1)       );

  //instantiate an adder to create the real part of the result
  add i_add_1 ( .clock (pipeclk),
                .dataa (sub_1),
                .datab (c_re_7q_se),
                .overflow (add_1_of),
                .result (add_1)       );

  //pare down the real part of the result to 16 bits
  assign result_re = {add_1[(`DIGITS * 2) - 1], add_1[`DIGITS - 2:0]};

  //instantiate an adder to create the imag part of the result
  add i_add_2 ( .clock (pipeclk),
                .dataa (mult_3_norm),
                .datab (c_im_5q_se),
                .overflow (add_2_of),
                .result (add_2)       );

  //truncate the result of the adder from 32 bits to 16 bits
  assign add_2_trunc = {add_2[(`DIGITS * 2) - 1], add_2[`DIGITS - 2:0]};



  //-----------------------
  //magnitude squared logic
  //-----------------------





  //instantiate an adder for z_re^2 + z_im^2.  this is the magnitude squared.
  add i_add_3 ( pipeclk,
                mult_1_norm,
                mult_2_norm,
                add_3_of,
                add_3         );




  //------------------------
  //general logic
  //------------------------




  //create a master overflow signal
  assign overflow = (add_1_of | of_b);

  //signal that the complex coordinate is done, either by overflowing (not in
  //  set) or by making it through the max number of iterations (in set).
  assign done = valid_q[`PIPE_STAGES - 1] & (iter_done | overflow);

  //this coordinate was in the set if done and no overflow
  assign in_set = (done & !overflow);




endmodule


