//-----------------------------------------------------------------------------
//  Mandelbrot Set Generator
//  ECE573 Project, Winter 2008
//  Author:  Jesse Armagost
//-----------------------------------------------------------------------------

`timescale 1 ps / 1 ps

module engine (
  input                                       pipe_reset_n,
  input                                       pipeclk,
  input                                       sysclk,
  //front end fifo interface
  input [`FE_FIFO_WIDTH - 1:0]                fe_fifo_wrdata,
  input                                       fe_fifo_wrreq,
  output                                      fe_fifo_wrfull,
  //visibility signals
  output [`NUM_PIPES - 1:0]                   cp_full,
  output [`NUM_PIPES - 1:0]                   be_fifo_wrfull,
  //back end fifo interface
  input [`NUM_PIPES - 1:0]                    be_fifo_rdreq,
  output [`NUM_PIPES * `BE_FIFO_WIDTH - 1:0]  be_fifo_rddata,
  output [`NUM_PIPES - 1:0]                   be_fifo_rdempty
);


  //regs and wires----------------------

  //front end fifo signals
  wire [`FE_FIFO_WIDTH - 1:0] fe_fifo_rddata;
  wire fe_fifo_rdreq;
  wire fe_fifo_rdempty;

  //complex pipe signals
  reg [`NUM_PIPES - 1:0] cp_load_en;
  wire [`NUM_PIPES - 1:0] cp_busy;
  wire [`NUM_PIPES - 1:0] cp_done;
  wire [`NUM_PIPES - 1:0] cp_in_set;
  wire [9:0] cp_v_pixel_in;
  wire [9:0] cp_h_pixel_in;
  wire [`DIGITS - 1:0] cp_c_re_in;
  wire [`DIGITS - 1:0] cp_c_im_in;
  wire [15:0] cp_iterations [`NUM_PIPES - 1:0];
  wire [9:0] cp_v_pixel_out [`NUM_PIPES - 1:0];
  wire [9:0] cp_h_pixel_out [`NUM_PIPES - 1:0];

  //back end fifo signals
  reg [`NUM_PIPES - 1:0] be_fifo_wrreq;
  wire [6:0] be_fifo_wrusedw [`NUM_PIPES - 1:0];
  reg [`BE_FIFO_WIDTH - 1:0] be_fifo_wrdata [`NUM_PIPES - 1:0];

  //general signals
  reg [`NUM_PIPES - 1:0] cp_load_inhibit;
  reg [`LB2_NUM_PIPES - 1:0] current_pipe;
  wire load_en;
  integer i;
  genvar j;

  //generate the complex pipe load inhibit signals.  These are asserted when
  //  1.  the current complex pipe slot is busy, or
  //  2.  the back end fifo for a complex pipe is past the high water mark, or
  //  3.  the back end fifo is full.
  //the backend high water mark is 120 out of 128 words, generated when the
  //  wrusedw[] signal has bits [6:3] asserted.  the back end fifo high water
  //  mark gives the fifo enough space to accept coordinates that may have
  //  already been in the complex pipe.
  always @(*) begin
    for (i = 0; i < `NUM_PIPES; i = i + 1) begin
      if (  (be_fifo_wrusedw[i][6:3] == 4'b1111)
          | be_fifo_wrfull[i]
          | cp_busy[i]                            )
        cp_load_inhibit[i] = 1'b1;
      else
        cp_load_inhibit[i] = 1'b0;
    end
  end

  //determine the pipe that currently has an empty slot.  in the case that none
  //  of the pipes have an empty slot, we still assign current_pipe to 0, but
  //  the pipe load_en signal will not be asserted so it doesn't matter.  we
  //  give lower numbered pipes priority by traversing the pipes from highest
  //  to lowest.
  always @(*) begin
    current_pipe = 0;
    for (i = `NUM_PIPES - 1; i >= 0; i = i - 1) begin
      if (~cp_load_inhibit[i])
        current_pipe = i[`LB2_NUM_PIPES - 1:0];
    end
  end

  //route the complex pipe load enable signal to the correct pipe.
  always @(*) begin
    for (i = 0; i < `NUM_PIPES; i = i + 1) begin
      if (i == current_pipe & load_en)
        cp_load_en[i] <= 1'b1;
      else
        cp_load_en[i] <= 1'b0;
    end
  end

  //pick the complex pipe input data out of the front end fifo read data.  the
  //  front end fifo bitfields are:
  //
  //    [zz:yy] = c_im
  //    [xx:20] = c_re
  //    [19:10] = v_pixel
  //    [9:0]   = h_pixel
  //
  assign cp_c_im_in = fe_fifo_rddata[19 + 2 * `DIGITS:20 + `DIGITS];
  assign cp_c_re_in = fe_fifo_rddata[19 + `DIGITS:20];
  assign cp_v_pixel_in = fe_fifo_rddata[19:10];
  assign cp_h_pixel_in = fe_fifo_rddata[9:0];

  //generate the complex pipe load enable signal if the front end fifo is not
  //  empty and if there is a complex pipe without its load inhibit signal
  //  asserted.
  assign load_en = (  ~fe_fifo_rdempty
                    & ~(&cp_load_inhibit) ) ? 1'b1 : 1'b0;

  //the front end fifo read acknowledge signal is just the complex pipe load_en
  //  signal.
  assign fe_fifo_rdreq = load_en;

  //the backend fifo write request signals are the done signals from the
  //  complex pipes.
  always @(*) begin
    for (i = 0; i < `NUM_PIPES; i = i + 1)
      be_fifo_wrreq[i] = cp_done[i];
  end

  //form the back end fifo write data for each fifo.
  //
  //    [36]    = in_set
  //    [35:20] = iterations
  //    [19:10] = v_pixel
  //    [9:0]   = h_pixel
  //
  always @(*) begin
    for (i = 0; i < `NUM_PIPES; i = i + 1) begin
      be_fifo_wrdata[i] <= {  cp_in_set[i],
                              cp_iterations[i],
                              cp_v_pixel_out[i],
                              cp_h_pixel_out[i]   };
    end
  end

  //instantiate the front end fifo.
  fe_fifo fe_fifo ( .data (fe_fifo_wrdata),
                    .rdclk (pipeclk),
                    .rdreq (fe_fifo_rdreq),
                    .wrclk (sysclk),
                    .wrreq (fe_fifo_wrreq),
                    .q (fe_fifo_rddata),
                    .rdempty (fe_fifo_rdempty),
                    .wrfull (fe_fifo_wrfull)    );

  //instantiate the complex pipes
  generate
    for (j = 0; j < `NUM_PIPES; j = j + 1) begin : cp
      complex_pipe cp ( .reset_n (pipe_reset_n),
                        .pipeclk (pipeclk),
                        .c_re_in (cp_c_re_in),
                        .c_im_in (cp_c_im_in),
                        .h_pixel_in (cp_h_pixel_in),
                        .v_pixel_in (cp_v_pixel_in),
                        .load_en (cp_load_en[j]),
                        .busy (cp_busy[j]),
                        .full (cp_full[j]),
                        .done (cp_done[j]),
                        .iterations (cp_iterations[j]),
                        .in_set (cp_in_set[j]),
                        .h_pixel_out (cp_h_pixel_out[j]),
                        .v_pixel_out (cp_v_pixel_out[j])  );
    end
  endgenerate

  //instantiate the back end fifos
  generate
    for (j = 0; j < `NUM_PIPES; j = j + 1) begin : be_fifo
      be_fifo be_fifo ( .data (be_fifo_wrdata[j]),
                        .rdclk (sysclk),
                        .rdreq (be_fifo_rdreq[j]),
                        .wrclk (pipeclk),
                        .wrreq (be_fifo_wrreq[j]),
                        .q (be_fifo_rddata[
                              `BE_FIFO_WIDTH - 1 + `BE_FIFO_WIDTH * j
                            : `BE_FIFO_WIDTH * j                      ]),
                        .rdempty (be_fifo_rdempty[j]),
                        .wrfull (be_fifo_wrfull[j]),
                        .wrusedw (be_fifo_wrusedw[j])   );
    end
  endgenerate




endmodule

