//-----------------------------------------------------------------------------
//  Mandelbrot Set Generator
//  ECE573 Project, Winter 2008
//  Author:  Jesse Armagost
//-----------------------------------------------------------------------------


`timescale 1 ps / 1 ps

`include "mand_defs.v"
`include "vga_defs_1024_768.v"

module mand (
  input                                           sys_reset_n,
  input                                           sysclk,
  input                                           pan_left,
  input                                           pan_right,
  input                                           pan_up,
  input                                           pan_down,
  input                                           zoom_in,
  input                                           zoom_out,
  input                                           start,
  input                                           restart,
  output      [9:0]                               current_line,
  output                                          rendering,
  output                                          frame_done,
  input                                           sram_wr_start_en,
  output                                          sram_wr_in_progress,
  output                                          zoom_in_en_led,
  output                                          uart_rx_en,
  //front end fifo interface
  output                                          fe_fifo_wrreq,
  output      [`FE_FIFO_WIDTH - 1:0]              fe_fifo_wrdata,
  input                                           fe_fifo_wrfull,
  //back end fifo interfaces
  output      [`NUM_PIPES - 1:0]                  be_fifo_rdreq,
  input       [`NUM_PIPES * `BE_FIFO_WIDTH - 1:0] be_fifo_rddata,
  input       [`NUM_PIPES - 1:0]                  be_fifo_rdempty,
  //back end sram interface
  output                                          sram_we_n,
  output      [17:0]                              sram_addr,
  output  reg [15:0]                              sram_wr_data,
  output                                          sram_oe_n,
  input       [15:0]                              sram_rd_data
);


  //regs and wires----------------------

  integer i;

  //synchronize vgaclk domain -> sysclk domain.
  reg sram_wr_start_en_1q;
  reg sram_wr_start_en_2q;

  //the point in the complex plane that lies at the center of the screen.
  reg signed [`DIGITS - 1:0] sgn_complex_center_re;
  reg signed [`DIGITS - 1:0] sgn_complex_center_im;

  //the point in the complex plane that lies at the upper left of the screen.
  //  this is where we begin rendering the complex plane.  we have to derive
  //  this point from the center point since we zoom in and out with respect to
  //  the center point.
  wire signed [`DIGITS - 1:0] sgn_complex_start_re;
  wire signed [`DIGITS - 1:0] sgn_complex_start_im;

  //the complex plane pixel increment.  this is how much we move horizontally
  //  or vertically in the complex plane when we move one screen pixel.  this
  //  value is used as we sweep through all the pixels on the screen.
  reg signed [`DIGITS - 1:0] sgn_complex_pixel_delta;
  wire signed [`DIGITS - 1:0] sgn_complex_pixel_delta_next;
  wire zoom_in_en;

  //the current complex plane coordinate.  these start at "sgn_complex_start_re"
  //  and "sgn_complex_start_im" and change as we work our way through all of the
  //  screen pixels.
  reg signed [`DIGITS - 1:0] sgn_current_coord_re;
  reg signed [`DIGITS - 1:0] sgn_current_coord_im;

  //the pixel counters.  these count through all of the horizontal and vertical
  //  screen pixels.  the control signals for these counters also control the
  //  complex plane current coordinate registers above, since they have to work
  //  closely together to provide the correct raster scanning action.
  reg [9:0] v_pixel_count;
  reg [9:0] h_pixel_count;
  wire pixel_count_clear;
  wire pixel_count_inc;
  wire v_pixel_count_done;
  wire h_pixel_count_done;

  //remember if this our first time running so we can render the starting frame
  //  without having to wait for the start signal.
  reg first_time;

  //this signal is used to auto-draw the frame if it's the first time we're
  //  running, or if the restart key was pressed.
  wire auto_start;

  //create a delayed version of restart, which will cause the frame to auto-draw
  reg restart_1q;
  reg restart_2q;

  //a master signal that indicates when all back end fifos are empty.
  wire be_fifos_empty;

  //the read request signal for the backend fifos, generated by the back end
  //  state machine.  this signal is routed to the appropriate back end fifo,
  //  determined by current_be_fifo.
  wire be_rdreq;

  //an internal version of the back end fifo read request signals.
  reg [`NUM_PIPES - 1:0] be_fifo_rdreq_int;

  //the backend fifo data is one large vector.  we need to extract the
  //  appropriate bits out of the read data depending on the current_be_fifo.
  //  this signal will hold that data.  each fifo's data is 44 bits wide.
  reg [`BE_FIFO_WIDTH - 1:0] current_be_fifo_rddata;

  //sram read data latch.  needed because we read from the sram and then alter
  //  the data and write it back, and we can't read and write at the same time.
  reg [15:0] sram_data_latch;
  wire sram_data_latch_en;

  //the current back end fifo that we are reading from.
  reg [`LB2_NUM_PIPES - 1:0] current_be_fifo;

  //we need to be able to latch the be_current_fifo signal because it could
  //  change as soon as we read from a back end fifo.  the latching is
  //  controlled by be_rdreq because when this signal is asserted we are reading
  //  from the current back end fifo.  after we latch be_current_fifo, we use it
  //  to determine which bit slice to grab out of the back end fifo read data.
  reg [`LB2_NUM_PIPES - 1:0] current_be_fifo_latch;

  //determine the pixel data based on whether the coordinate is in the set and
  //  the number of iterations.  this is the data that is written to the sram.
  wire [3:0] pixel_data;

  //an internal version of frame done, used by the state machine.
  wire frame_done_int;

  //the front end state machine.  this machine idles until it receives the
  //  start signal, at which point it works its way through all the screen
  //  pixels (and complex plane coordinates) and writes them to the front end
  //  fifo.  if this is our first time running, indicated by first_time, then
  //  this machine automatically runs.
  reg [2:0] fe_current_state;
  reg [2:0] fe_next_state;

  parameter fe_s_idle =     3'b001;   //wait for start
  parameter fe_s_assign =   3'b010;   //write to front end fifo or spin
  parameter fe_s_done =     3'b100;   //signal done

  //the back end state machine.  this machine reads from the complex pipe
  //  fifos and writes to the sram during the horizontal blanking periods.
  reg [6:0] be_current_state;
  reg [6:0] be_next_state;

  parameter be_s_idle =       7'b0000001;
  parameter be_s_sram_rd_1 =  7'b0000010;
  parameter be_s_sram_rd_2 =  7'b0000100;
  parameter be_s_sram_rd_3 =  7'b0001000;
  parameter be_s_sram_wr_1 =  7'b0010000;
  parameter be_s_sram_wr_2 =  7'b0100000;
  parameter be_s_sram_wr_3 =  7'b1000000;


//--------------------------------


  //remember if this is our first time running so we can auto-render the frame
  //  without waiting for the start signal.
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n)
      first_time <= 1'b1;
    else
      first_time <= 1'b0;
  end

  //create a delayed auto-start signal that will kick off rendering two cycles
  //  after the restart key was pressed.
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n) begin
      restart_1q <= 1'b0;
      restart_2q <= 1'b0;
    end
    else if (sysclk) begin
      restart_1q <= restart;
      restart_2q <= restart_1q;
    end
  end

  //we auto-start when it's the first time running or if the restart key was
  //  pressed one cycle ago.
  assign auto_start = (first_time | restart_2q);

  //adjust the complex plane center point when we pan left, right, up or down.
  //  when we pan, we move by the amount (`PAN_PIXELS * sgn_complex_pixel_delta).
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n) begin
      sgn_complex_center_re <= `INITIAL_COMPLEX_CENTER_RE;
      sgn_complex_center_im <= `INITIAL_COMPLEX_CENTER_IM;
    end
    else if (sysclk) begin
      if (restart) begin
        sgn_complex_center_re <= `INITIAL_COMPLEX_CENTER_RE;
        sgn_complex_center_im <= `INITIAL_COMPLEX_CENTER_IM;
      end
      if (pan_left) begin
        sgn_complex_center_re <=
              sgn_complex_center_re - (`PAN_PIXELS * sgn_complex_pixel_delta);
      end
      if (pan_right) begin
        sgn_complex_center_re <=
              sgn_complex_center_re + (`PAN_PIXELS * sgn_complex_pixel_delta);
      end
      if (pan_up) begin
        sgn_complex_center_im <=
              sgn_complex_center_im - (`PAN_PIXELS * sgn_complex_pixel_delta);
      end
      if (pan_down) begin
        sgn_complex_center_im <=
              sgn_complex_center_im + (`PAN_PIXELS * sgn_complex_pixel_delta);
      end
    end
  end

  //adjust the complex plane pixel delta when we zoom in or out.  zoom_in_en
  //  keeps us from zooming in too far.
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n) begin
      sgn_complex_pixel_delta <= `INITIAL_COMPLEX_PIXEL_DELTA;
    end
    else if (sysclk) begin
      if (restart) begin
        sgn_complex_pixel_delta <= `INITIAL_COMPLEX_PIXEL_DELTA;
      end
      if (zoom_in & zoom_in_en) begin
        sgn_complex_pixel_delta <= sgn_complex_pixel_delta_next;
      end
      if (zoom_out) begin
        sgn_complex_pixel_delta <= sgn_complex_pixel_delta <<< 1;
      end
    end
  end

  //predetermine the next zoom delta so we can prevent zooming if it will zoom
  //  us in too far.
  assign sgn_complex_pixel_delta_next = sgn_complex_pixel_delta >>> 1;

  assign zoom_in_en = (sgn_complex_pixel_delta_next == `DIGITS'sd0) ? 1'b0 : 1'b1;

  assign zoom_in_en_led = ~zoom_in_en;

  //derive the complex plane upper left starting point from the complex plane
  //  center point and the complex pixel delta.  we move horizontally half the
  //  total number of pixels times the complex pixel delta.  we move vertically
  //  half the total number of pixel times the complex pixel delta.
  assign sgn_complex_start_re =
              sgn_complex_center_re - `DIGITS'sd512 * sgn_complex_pixel_delta;
  assign sgn_complex_start_im =
              sgn_complex_center_im - `DIGITS'sd384 * sgn_complex_pixel_delta;

  //the horizontal and vertical pixel counters, and the current complex plane
  //  coordinate.  these are combined here because they work so closely
  //  together to provide the raster scanning effect.
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n) begin
      h_pixel_count <= 10'd0;
      v_pixel_count <= 10'd0;
      sgn_current_coord_re <= `DIGITS'sd0;
      sgn_current_coord_im <= `DIGITS'sd0;
    end
    else if (sysclk) begin
      if (pixel_count_clear) begin
        h_pixel_count <= 10'd0;
        v_pixel_count <= 10'd0;
        sgn_current_coord_re <= sgn_complex_start_re;
        sgn_current_coord_im <= sgn_complex_start_im;
      end
      else if (pixel_count_inc) begin
        if (h_pixel_count_done) begin
            v_pixel_count <= v_pixel_count + 1'b1;
            sgn_current_coord_im <=
                                sgn_current_coord_im + sgn_complex_pixel_delta;
          h_pixel_count <= 10'b0;
          sgn_current_coord_re <= sgn_complex_start_re;
        end
        else begin
          h_pixel_count <= h_pixel_count + 1'b1;
          sgn_current_coord_re <=
                                sgn_current_coord_re + sgn_complex_pixel_delta;
        end
      end
    end
  end

  assign h_pixel_count_done = (h_pixel_count == `H_PIXELS - 1) ? 1'b1 : 1'b0;
  assign v_pixel_count_done = (v_pixel_count == `V_PIXELS - 1) ? 1'b1 : 1'b0;
  assign frame_done_int = (h_pixel_count_done & v_pixel_count_done);

  assign current_line = v_pixel_count;



  //---------------------------------
  //front end stuff





  //create the front end fifo write data.
  //
  //  [zz:yy] = sgn_current_coord_im
  //  [xx:20] = sgn_current_coord_re
  //  [19:10] = v_pixel_count
  //  [9:0]   = h_pixel_count
  //
  assign fe_fifo_wrdata = { sgn_current_coord_im,
                            sgn_current_coord_re,
                            v_pixel_count,
                            h_pixel_count         };

  //front end state machine current state
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n)
      fe_current_state <= fe_s_idle;
    else if (sysclk)
      fe_current_state <= fe_next_state;
  end

  //front end state machine next state
  always @(*) begin
    case (fe_current_state)

      fe_s_idle : begin
        if (start | auto_start)
          fe_next_state <= fe_s_assign;
        else
          fe_next_state <= fe_s_idle;
      end

      fe_s_assign : begin
        if (frame_done_int)
          fe_next_state <= fe_s_done;
        else
          fe_next_state <= fe_s_assign;
      end

      fe_s_done : begin
        fe_next_state <= fe_s_idle;
      end

      default : begin
        fe_next_state <= fe_s_idle;
      end

    endcase
  end

  //front end state machine output logic
  assign pixel_count_clear = (    fe_current_state == fe_s_idle
                              &&  (start | auto_start)          ) ? 1'b1 : 1'b0;

  assign uart_rx_en = (fe_current_state == fe_s_idle);

  assign pixel_count_inc = (    fe_current_state == fe_s_assign
                            &&  ~fe_fifo_wrfull                 ) ? 1'b1 : 1'b0;

  assign fe_fifo_wrreq = (    fe_current_state == fe_s_assign
                          &&  ~fe_fifo_wrfull                 ) ? 1'b1 : 1'b0;

  assign rendering = (fe_current_state == fe_s_idle) ? 1'b0 : 1'b1;

  assign frame_done = (fe_current_state == fe_s_done) ? 1'b1 : 1'b0;


  //---------------------------------
  //back end stuff





  //create a master signal that indicates when all back end fifos are empty.
  assign be_fifos_empty = (&be_fifo_rdempty);

  //determine the current back end fifo to read from.  by traversing the pipes
  //  from lowest to highest, we give priority to the highest fifo that is not
  //  empty.
  always @(*) begin
    current_be_fifo = 0;
    for (i = 0; i < `NUM_PIPES; i = i + 1) begin
      if (~be_fifo_rdempty[i])
        current_be_fifo = i[`LB2_NUM_PIPES - 1:0];
    end
  end

  //pipe the back end read request to the appropriate back end fifo.
  always @(*) begin
    be_fifo_rdreq_int = {`NUM_PIPES{1'b0}};
    for (i = 0; i < `NUM_PIPES; i = i + 1) begin
      if (    current_be_fifo == i
          &&  be_rdreq              )
        be_fifo_rdreq_int[i] = 1'b1;
      else
        be_fifo_rdreq_int[i] = 1'b0;
    end
  end

  assign be_fifo_rdreq = be_fifo_rdreq_int;

  //implement the sram read data latch for when we perform our read-modify-write
  //  cycles.
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n)
      sram_data_latch <= 16'h0000;
    else if (sysclk) begin
      if (sram_data_latch_en)
        sram_data_latch <= sram_rd_data;
    end
  end

  //form the address to drive out to the sram.  the upper bits are the vertical
  //  pixel value and the lower bits are the horizontal pixel value.  the 2 lsbs
  //  of the horizontal pixel value aren't used for the address because they
  //  select the correct nibble out of the ram data.
  assign sram_addr = {  current_be_fifo_rddata[19:10],
                        current_be_fifo_rddata[9:2]    };

  //latch the current back end fifo number.  the latching is controlled by the
  //  back end fifo read request signal.
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n)
      current_be_fifo_latch <= {`LB2_NUM_PIPES{1'b0}};
    else if (sysclk) begin
      if (be_rdreq)
        current_be_fifo_latch <= current_be_fifo;
    end
  end

  //extract the correct read data out of the one large vector of back end fifo
  //  read data.  we need to use the latched version of current_be_fifo because
  //  the original version of the signal could change after we issue the fifo
  //  read signal.  the actual back end fifo data stays constant until we assert
  //  its rdreq signal again.
  always @(*) begin
    case (current_be_fifo_latch)
//       3'd7 : current_be_fifo_rddata <=
//                 be_fifo_rddata[`BE_FIFO_WIDTH * 8 - 1:`BE_FIFO_WIDTH * 7];
//       3'd6 : current_be_fifo_rddata <=
//                 be_fifo_rddata[`BE_FIFO_WIDTH * 7 - 1:`BE_FIFO_WIDTH * 6];
//       3'd5 : current_be_fifo_rddata <=
//                 be_fifo_rddata[`BE_FIFO_WIDTH * 6 - 1:`BE_FIFO_WIDTH * 5];
//       3'd4 : current_be_fifo_rddata <=
//                 be_fifo_rddata[`BE_FIFO_WIDTH * 5 - 1:`BE_FIFO_WIDTH * 4];
//       3'd3 : current_be_fifo_rddata <=
//                 be_fifo_rddata[`BE_FIFO_WIDTH * 4 - 1:`BE_FIFO_WIDTH * 3];
//       3'd2 : current_be_fifo_rddata <=
//                 be_fifo_rddata[`BE_FIFO_WIDTH * 3 - 1:`BE_FIFO_WIDTH * 2];
      3'd1 : current_be_fifo_rddata <=
                be_fifo_rddata[`BE_FIFO_WIDTH * 2 - 1:`BE_FIFO_WIDTH    ];
      default : current_be_fifo_rddata <=
                be_fifo_rddata[`BE_FIFO_WIDTH - 1    :0                 ];
    endcase
  end

  //determine pixel data to be stored in the sram, based on in_set.
  //assign pixel_data = (current_be_fifo_rddata[36]) ? 4'b1111 : 4'b0000;
  assign pixel_data = current_be_fifo_rddata[36] ? 4'b0000
                    : current_be_fifo_rddata[26:23];

  //synchronize the signal coming from the vga clock domain
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n) begin
      sram_wr_start_en_1q <= 0;
      sram_wr_start_en_2q <= 0;
    end
    else if (sysclk) begin
      sram_wr_start_en_1q <= sram_wr_start_en;
      sram_wr_start_en_2q <= sram_wr_start_en_1q;
    end
  end

  //drive the appropriate sram write data out.  we alter the correct nibble
  //  of the sram data depending on the low two bits of the horizontal pixel,
  //  which is located in the data we read from the back end fifo.
  //
  always @(*) begin
    case (current_be_fifo_rddata[1:0])
      //low byte, low nibble
      2'b00 : begin
        sram_wr_data <= { sram_data_latch[15:4],
                          pixel_data              };
      end
      //low byte, high nibble
      2'b01 : begin
        sram_wr_data <= { sram_data_latch[15:8],
                          pixel_data,
                          sram_data_latch[3:0]    };
      end
      //high byte, low nibble
      2'b10 : begin
        sram_wr_data <= { sram_data_latch[15:12],
                          pixel_data,
                          sram_data_latch[7:0]    };
      end
      //high byte, high nibble
      default : begin
        sram_wr_data <= { pixel_data,
                          sram_data_latch[11:0]   };
      end
    endcase
  end

  //back end state machine current state
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n)
      be_current_state <= be_s_idle;
    else if (sysclk)
      be_current_state <= be_next_state;
  end

  //back end state machine next state
  always @(*) begin
    case (be_current_state)
      be_s_idle : begin
        if (  ~be_fifos_empty
            & sram_wr_start_en_2q )
          be_next_state <= be_s_sram_rd_1;
        else
          be_next_state <= be_s_idle;
      end
      be_s_sram_rd_1 : begin
        be_next_state <= be_s_sram_rd_2;
      end
      be_s_sram_rd_2 : begin
        be_next_state <= be_s_sram_rd_3;
      end
      be_s_sram_rd_3 : begin
        be_next_state <= be_s_sram_wr_1;
      end
      be_s_sram_wr_1 : begin
        be_next_state <= be_s_sram_wr_2;
      end
      be_s_sram_wr_2 : begin
        be_next_state <= be_s_sram_wr_3;
      end
      be_s_sram_wr_3 : begin
        be_next_state <= be_s_idle;
      end
      default : begin
        be_next_state <= be_s_idle;
      end
    endcase
  end

  //back end state machine output logic
  assign be_rdreq = (   be_current_state == be_s_idle
                      & ~be_fifos_empty
                      & sram_wr_start_en_2q            ) ? 1'b1 : 1'b0;

  assign sram_oe_n = (    be_current_state == be_s_sram_rd_1
                      ||  be_current_state == be_s_sram_rd_2
                      ||  be_current_state == be_s_sram_rd_3  ) ? 1'b0 : 1'b1;

  assign sram_we_n = (    be_current_state == be_s_sram_wr_1
                      ||  be_current_state == be_s_sram_wr_2  ) ? 1'b0 : 1'b1;

  assign sram_data_latch_en = (be_current_state == be_s_sram_rd_3) ?
                                                                  1'b1 : 1'b0;

  //the back end state machine is only allowed to initiate an sram write during
  //  the vga horizontal sync period, but if it starts a write at the very end
  //  the write may be truncated.  therefore, we need to have the machine
  //  indicate when it is running so it can have access to the sram a little
  //  past the vga horizontal sync period if need be.
  assign sram_wr_in_progress = (be_current_state == be_s_idle) ? 1'b0 : 1'b1;



endmodule
