//-----------------------------------------------------------------------------
//  Mandelbrot Set Generator
//  ECE573 Project, Winter 2008
//  Author:  Jesse Armagost
//-----------------------------------------------------------------------------

`timescale 1 ps / 1 ps

module sram_io (  input               sys_reset_n,
                  input               sysclk,
                  input               vga_reset_n,
                  input               vgaclk,
                  input               mand_sram_wr_in_progress,
                  input       [17:0]  mand_sram_addr,
                  input               mand_sram_oe_n,
                  input               mand_sram_we_n,
                  input       [15:0]  mand_sram_wr_data,
                  input       [17:0]  vga_sram_addr,
                  output      [17:0]  sram_addr,
                  output              sram_oe_n,
                  output reg          sram_we_n,
                  inout       [15:0]  sram_data,
                  output              sram_ce_n,
                  output              sram_lb_n,
                  output              sram_ub_n             );


  //regs and wires
  reg [17:0] vga_sram_addr_1q;
  reg [17:0] mand_sram_addr_1q;
  reg mand_sram_oe_n_1q;
  reg mand_sram_we_n_1q;
  reg [15:0] mand_sram_wr_data_1q;
  reg mand_sram_wr_in_progress_1q;

  //register the incoming signals from the mand module to help with timing.
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n) begin
      mand_sram_addr_1q <= 0;
      mand_sram_oe_n_1q <= 0;
      mand_sram_we_n_1q <= 0;
      mand_sram_wr_data_1q <= 0;
      mand_sram_wr_in_progress_1q <= 0;
    end
    else if (sysclk) begin
      mand_sram_addr_1q <= mand_sram_addr;
      mand_sram_oe_n_1q <= mand_sram_oe_n;
      mand_sram_we_n_1q <= mand_sram_we_n;
      mand_sram_wr_data_1q <= mand_sram_wr_data;
      mand_sram_wr_in_progress_1q <= mand_sram_wr_in_progress;
    end
  end

  //register the incoming signals from the vga controller module to help with
  //  timing.
  always @(negedge vga_reset_n or posedge vgaclk) begin
    if (~vga_reset_n)
      vga_sram_addr_1q <= 0;
    else if (vgaclk)
      vga_sram_addr_1q <= vga_sram_addr;
  end

  //mux the sram address out
  assign sram_addr = (mand_sram_wr_in_progress_1q)  ? mand_sram_addr_1q
                                                    : vga_sram_addr_1q;

  //mux the oe signal out
  assign sram_oe_n = (mand_sram_wr_in_progress_1q)  ? mand_sram_oe_n_1q
                                                    : 1'b0;

  //control the tristate data out buffer
  assign sram_data = (  mand_sram_wr_in_progress_1q
                      & sram_oe_n                   ) ? mand_sram_wr_data_1q
                                                      : 16'hzzzz;

  //mux the we signal out.  we have to register write enable so we don't
  //  accidentally produce writes when the sram access mux switches back and
  //  forth.  this means that we've got to hold the other sram signals for
  //  longer than expected in the mand module, where we are reading and
  //  writing pixel data from/to the sram.
  always @(negedge sys_reset_n or posedge sysclk) begin
    if (~sys_reset_n)
      sram_we_n <= 1'b1;
    else if (sysclk)
      sram_we_n <= (mand_sram_wr_in_progress_1q) ? mand_sram_we_n_1q : 1'b1;
  end

  //always enable the sram chip and both byte enables
  assign sram_ce_n = 1'b0;
  assign sram_lb_n = 1'b0;
  assign sram_ub_n = 1'b0;



endmodule

