//-----------------------------------------------------------------------------
//  Mandelbrot Set Generator
//  ECE573 Project, Winter 2008
//  Authors:  Jesse Armagost and Eddie Yang
//-----------------------------------------------------------------------------



//------------------------------------------------------------------------------
//    In the receive direction,
//    full autobaud capability is provided to autobaud to a speed between 9600
//    and 38400.  We must autobaud before we can use the transmit side of the
//    UART.  A good autobaud character is the return key, since its start bit
//    appears alone.
//
//------------------------------------------------------------------------------

`timescale 1 ps / 1 ps

module uart ( input reset_n,
              input clk,
              input rx_en,                      //from mand, disables input
                                                //  while mand is running
              input echo,                       //response from uart_decode
              input uart_rx,                    //bit in
              output reg uart_tx,               //bit out
              output reg [7:0] rx_data_reg,     //data from uart rx
              output reg rx_rdy);               //data from uart is ready


  //regs and wires
  reg [3:0] clock_en_count;       //generate a slow clock enable
  wire clock_en;
  reg bit_time_known;             //have we autobauded yet?
  reg bit_time_known_set_en;
  wire [7:0] bit_time_half;       //half a bit time
  wire [8:0] bit_time_whole;      //a whole bit time
  wire [12:0] bit_time_sixteen;   //sixteen bit times

  //rx side
  reg [8:0] rx_bit_time_count;    //count length of a bit
  reg rx_bit_time_count_en;
  reg [12:0] rx_bit_delay_count;
  reg rx_bit_delay_count_clear;
  reg rx_bit_delay_count_inc;
  wire rx_bit_delay_half_done;
  wire rx_bit_delay_whole_done;
  wire rx_bit_delay_sixteen_done;
  reg [2:0] rx_bit_count;         //count bits in a byte
  reg rx_bit_count_clear;
  wire rx_bit_count_done;
  reg rx_latch_bit_en;            //shift the new bit in
  reg uart_rx_1q;
  reg uart_rx_2q;
  reg clear_rx_flag;

  reg [8:0] rx_current_state;
  reg [8:0] rx_next_state;
  parameter s_rx_idle =              9'h001;
  parameter s_rx_measure =           9'h002;
  parameter s_rx_wait =              9'h004;
  parameter s_rx_wait_char =         9'h008;
  parameter s_rx_wait_half_bit =     9'h010;
  parameter s_rx_wait_whole_bit_a =  9'h020;
  parameter s_rx_latch_bit =         9'h040;
  parameter s_rx_wait_whole_bit_b =  9'h080;
  parameter s_rx_set_flag =          9'h100;

  //tx side
  reg [12:0] tx_bit_delay_count;
  reg tx_bit_delay_count_clear;
  reg tx_bit_delay_count_inc;
  wire tx_bit_delay_whole_done;
  reg [3:0] tx_bit_count;           //count start bit, data, stop bit
  reg tx_bit_count_clear;
  wire tx_bit_count_done;
  reg [13:0] tx_data_reg;           //assemble the incoming byte
  reg tx_data_reg_rotate_en;
  wire uart_tx_1q;

  reg [3:0] tx_current_state;
  reg [3:0] tx_next_state;
  parameter s_tx_idle =       4'h1;
  parameter s_tx_wait_char =  4'h2;
  parameter s_tx_send_bit =   4'h4;
  parameter s_tx_rotate =     4'h8;


  //---------------------------------------------------------------------------
  //implement the "bit_time_known" flag.  this flag is set after we autobaud.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      bit_time_known <= 0;
    else if (clk) begin
      if (bit_time_known_set_en)
        bit_time_known <= 1'b1;
    end
  end

  //---------------------------------------------------------------------------
  //implement the register that is used to assemble an rx byte.  incoming bits
  //  are shifted in the msb end because lsb is transmitted first on rs232.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      rx_data_reg <= 0;
    else if (clk) begin
      if (rx_latch_bit_en)
        rx_data_reg <= {uart_rx_2q, rx_data_reg[7:1]};
    end
  end

  //---------------------------------------------------------------------------
  //count bits in an incoming byte.  we ignore the start bit and count 8 data
  //  bits.  we brazenly ignore the stop bit, too.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      rx_bit_count <= 0;
    else if (clk) begin
      if (rx_bit_count_clear)
        rx_bit_count <= 3'b000;
      else if (rx_latch_bit_en)
        rx_bit_count <= rx_bit_count + 1'b1;
    end
  end

  assign rx_bit_count_done = (rx_bit_count == 4'b111 ? 1'b1 : 1'b0);

  //---------------------------------------------------------------------------
  //synchronize the incoming rx bit.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n) begin
      uart_rx_1q <= 1'b1;
      uart_rx_2q <= 1'b1;
    end
    else if (clk) begin
      uart_rx_2q <= uart_rx_1q;
      uart_rx_1q <= uart_rx;
    end
  end

  //---------------------------------------------------------------------------
  //create clock enable with period of xxx ns.  This corresponds to roughly
  //  xx samples per bit at our highest rate of 115200.  At our lowest rate of
  //  9600, this corresponds to xxx samples.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      clock_en_count <= 0;
    else if (clk) begin
      clock_en_count <= clock_en_count + 1'b1;
    end
  end

  assign clock_en = (clock_en_count == 4'b1111) ? 1'b1 : 1'b0;

  //---------------------------------------------------------------------------
  //create the bit time counter.  we wait for the first start bit and then
  //  enable this counter for the whole time the start bit is low.  of course
  //  the received character has to be one with the first data bit high, so we
  //  can measure a single bit time.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      rx_bit_time_count <= 0;
    else if (clk & clock_en) begin
      if (rx_bit_time_count_en)
        rx_bit_time_count <= rx_bit_time_count + 1'b1;
    end
  end

  assign bit_time_half = rx_bit_time_count[8:1];
  assign bit_time_whole = rx_bit_time_count;
  assign bit_time_sixteen = {rx_bit_time_count, 4'b0000};

  //---------------------------------------------------------------------------
  //implement the bit delay counter.  this counter is used to wait for
  //  half a bit time, a whole bit time or 16 bit times.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      rx_bit_delay_count <= 0;
    else if (clk) begin
      if (rx_bit_delay_count_clear)
        rx_bit_delay_count <= 0;
      else if (clock_en & rx_bit_delay_count_inc)
        rx_bit_delay_count <= rx_bit_delay_count + 1'b1;
    end
  end

  assign rx_bit_delay_half_done =
                      (rx_bit_delay_count[7:0] == bit_time_half ? 1'b1 : 1'b0);

  assign rx_bit_delay_whole_done =
                    (rx_bit_delay_count[8:0] == bit_time_whole ? 1'b1 : 1'b0);

  assign rx_bit_delay_sixteen_done =
                        (rx_bit_delay_count == bit_time_sixteen ? 1'b1 : 1'b0);

  //---------------------------------------------------------------------------
  //rx state machine current state logic
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      rx_current_state <= s_rx_idle;
    else if (clk)
      rx_current_state <= rx_next_state;
  end

  //---------------------------------------------------------------------------
  //rx state machine next state logic
  //---------------------------------------------------------------------------
  always @(*) begin
    case (rx_current_state)

      s_rx_idle: begin
        if ((uart_rx_2q == 1'b0))
          rx_next_state <= s_rx_measure;
        else
          rx_next_state <= s_rx_idle;
      end

      s_rx_measure: begin
        if (uart_rx_2q == 1'b1)
          rx_next_state <= s_rx_wait;
        else
          rx_next_state <= s_rx_measure;
      end

      s_rx_wait: begin
        if (rx_bit_delay_sixteen_done)
          rx_next_state <= s_rx_wait_char;
        else
          rx_next_state <= s_rx_wait;
      end

      s_rx_wait_char: begin
        if ((uart_rx_2q == 1'b0) & rx_en)
          rx_next_state <= s_rx_wait_half_bit;
        else
          rx_next_state <= s_rx_wait_char;
      end

      s_rx_wait_half_bit: begin
        if (rx_bit_delay_half_done)
          rx_next_state <= s_rx_wait_whole_bit_a;
        else
          rx_next_state <= s_rx_wait_half_bit;
      end

      s_rx_wait_whole_bit_a: begin
        if (rx_bit_delay_whole_done)
          rx_next_state <= s_rx_latch_bit;
        else
          rx_next_state <= s_rx_wait_whole_bit_a;
      end

      s_rx_latch_bit: begin
        if (rx_bit_count_done)
          rx_next_state <= s_rx_wait_whole_bit_b;
        else
          rx_next_state <= s_rx_wait_whole_bit_a;
      end

      s_rx_wait_whole_bit_b: begin
        if (rx_bit_delay_whole_done)
          rx_next_state <= s_rx_set_flag;
        else
          rx_next_state <= s_rx_wait_whole_bit_b;
      end

      s_rx_set_flag: begin
        rx_next_state <= s_rx_wait_char;
      end

      default: begin
        rx_next_state <= s_rx_idle;
      end

    endcase
  end

  //---------------------------------------------------------------------------
  //rx state machine output logic
  //---------------------------------------------------------------------------
  always @(*) begin
    case (rx_current_state)

      s_rx_idle: begin
        rx_bit_time_count_en <= 1'b0;
        rx_bit_delay_count_clear <= 1'b0;
        rx_bit_delay_count_inc <= 1'b0;
        rx_latch_bit_en <= 1'b0;
        rx_rdy <= 1'b0;
        rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b0;
      end

      s_rx_measure: begin
        rx_bit_time_count_en <= 1'b1;
        if (rx_next_state == s_rx_wait)
          rx_bit_delay_count_clear <= 1'b1;
        else
          rx_bit_delay_count_clear <= 1'b0;
        rx_bit_delay_count_inc <= 1'b0;
        rx_latch_bit_en <= 1'b0;
        rx_rdy <= 1'b0;
        rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b0;
      end

      s_rx_wait: begin
        rx_bit_time_count_en <= 1'b0;
        rx_bit_delay_count_clear <= 1'b0;
        rx_bit_delay_count_inc <= 1'b1;
        rx_latch_bit_en <= 1'b0;
        rx_rdy <= 1'b0;
        rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b0;
      end

      s_rx_wait_char: begin
        rx_bit_time_count_en <= 1'b0;
        if (rx_next_state == s_rx_wait_half_bit)
          rx_bit_delay_count_clear <= 1'b1;
        else
          rx_bit_delay_count_clear <= 1'b0;
        rx_bit_delay_count_inc <= 1'b0;
        rx_latch_bit_en <= 1'b0;
        rx_rdy <= 1'b0;
        if (rx_next_state == s_rx_wait_half_bit)
          rx_bit_count_clear <= 1'b1;
        else
          rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b1;
      end

      s_rx_wait_half_bit: begin
        rx_bit_time_count_en <= 1'b0;
        if (rx_next_state == s_rx_wait_whole_bit_a)
          rx_bit_delay_count_clear <= 1'b1;
        else
          rx_bit_delay_count_clear <= 1'b0;
        rx_bit_delay_count_inc <= 1'b1;
        rx_latch_bit_en <= 1'b0;
        rx_rdy <= 1'b0;
        rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b1;
      end

      s_rx_wait_whole_bit_a: begin
        rx_bit_time_count_en <= 1'b0;
        rx_bit_delay_count_clear <= 1'b0;
        rx_bit_delay_count_inc <= 1'b1;
        rx_latch_bit_en <= 1'b0;
        rx_rdy <= 1'b0;
        rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b1;
      end

      s_rx_latch_bit: begin
        rx_bit_time_count_en <= 1'b0;
        rx_bit_delay_count_clear <= 1'b1;
        rx_bit_delay_count_inc <= 1'b0;
        rx_latch_bit_en <= 1'b1;
        rx_rdy <= 1'b0;
        rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b1;
      end

      s_rx_wait_whole_bit_b: begin
        rx_bit_time_count_en <= 1'b0;
        rx_bit_delay_count_clear <= 1'b0;
        rx_bit_delay_count_inc <= 1'b1;
        rx_latch_bit_en <= 1'b0;
        rx_rdy <= 1'b0;
        rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b1;
      end

      s_rx_set_flag: begin
        rx_bit_time_count_en <= 1'b0;
        rx_bit_delay_count_clear <= 1'b0;
        rx_bit_delay_count_inc <= 1'b0;
        rx_latch_bit_en <= 1'b0;
        rx_rdy <= 1'b1;
        rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b1;
      end

      default: begin
        rx_bit_time_count_en <= 1'b0;
        rx_bit_delay_count_clear <= 1'b0;
        rx_bit_delay_count_inc <= 1'b0;
        rx_latch_bit_en <= 1'b0;
        rx_rdy <= 1'b0;
        rx_bit_count_clear <= 1'b0;
        bit_time_known_set_en <= 1'b0;
      end

    endcase
  end


  //---------------------------------------------------------------------------

  //---------------------------------------------------------------------------


  //---------------------------------------------------------------------------
  //implement the tx bit delay counter
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      tx_bit_delay_count <= 0;
    else if (clk) begin
      if (tx_bit_delay_count_clear)
        tx_bit_delay_count <= 0;
      else if (clock_en & tx_bit_delay_count_inc)
        tx_bit_delay_count <= tx_bit_delay_count + 1'b1;
    end
  end

  assign tx_bit_delay_whole_done =
                    (tx_bit_delay_count[8:0] == bit_time_whole ? 1'b1 : 1'b0);

  //---------------------------------------------------------------------------
  //implement the tx bit counter to count transmitted bits.  we need to count
  //  to 10, since we add a start bit and a done bit.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      tx_bit_count <= 0;
    else if (clk) begin
      if (tx_bit_count_clear)
        tx_bit_count <= 0;
      else if (tx_data_reg_rotate_en)
        tx_bit_count <= tx_bit_count + 1'b1;
    end
  end

  assign tx_bit_count_done = (tx_bit_count == 4'b1100 ? 1'b1 : 1'b0);

  //---------------------------------------------------------------------------
  //implement the tx transmit register.  we latch the tx data into this
  //  register, while adding a start and several stop bits (to not let the
  //  transmitter jam stuff out too fast).  then the register rotates as we
  //  transmit bits.
  //force the upper bit to zero because the pdp will set it, based on its
  //  notion of ascii.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      tx_data_reg <= 14'h3fff;
    else if (clk) begin
      if (bit_time_known) begin
        if (    tx_current_state == s_tx_wait_char
            &&  echo                          )
          tx_data_reg <= {5'b11111, 1'b0, rx_data_reg[6:0], 1'b0}; //stops, data, start
        else if (tx_data_reg_rotate_en)
          tx_data_reg <= {tx_data_reg[0], tx_data_reg[13:1]};
      end
    end
  end

  //---------------------------------------------------------------------------
  //register the outgoing tx data.
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      uart_tx <= 1'b1;
    else if (clk)
      uart_tx <= uart_tx_1q;
  end

  assign uart_tx_1q = tx_data_reg[0];

  //---------------------------------------------------------------------------
  //tx state machine current state logic
  //---------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      tx_current_state <= s_tx_idle;
    else if (clk)
      tx_current_state <= tx_next_state;
  end

  //---------------------------------------------------------------------------
  //tx state machine next state logic
  //---------------------------------------------------------------------------
  always @(*) begin
    case (tx_current_state)

      s_tx_idle: begin
        if (bit_time_known)
          tx_next_state <= s_tx_wait_char;
        else
          tx_next_state <= s_tx_idle;
      end

      s_tx_wait_char: begin
        if (echo)
          tx_next_state <= s_tx_send_bit;
        else
          tx_next_state <= s_tx_wait_char;
      end

      s_tx_send_bit: begin
        if (tx_bit_delay_whole_done)
          tx_next_state <= s_tx_rotate;
        else
          tx_next_state <= s_tx_send_bit;
      end

      s_tx_rotate: begin
        if (tx_bit_count_done)
          tx_next_state <= s_tx_wait_char;
        else
          tx_next_state <= s_tx_send_bit;
      end

      default: begin
        tx_next_state <= s_tx_idle;
      end

    endcase
  end

  //---------------------------------------------------------------------------
  //tx state machine output logic
  //---------------------------------------------------------------------------
  always @(*) begin
    case (tx_current_state)

      s_tx_idle: begin
        tx_bit_count_clear <= 1'b0;
        tx_bit_delay_count_clear <= 1'b0;
        tx_bit_delay_count_inc <= 1'b0;
        tx_data_reg_rotate_en <= 1'b0;
      end

      s_tx_wait_char: begin
        if (tx_next_state == s_tx_send_bit) begin
          tx_bit_count_clear <= 1'b1;
          tx_bit_delay_count_clear <= 1'b1;
        end
        else begin
          tx_bit_count_clear <= 1'b0;
          tx_bit_delay_count_clear <= 1'b0;
        end
        tx_bit_delay_count_inc <= 1'b0;
        tx_data_reg_rotate_en <= 1'b0;
      end

      s_tx_send_bit: begin
        tx_bit_count_clear <= 1'b0;
        tx_bit_delay_count_clear <= 1'b0;
        tx_bit_delay_count_inc <= 1'b1;
        tx_data_reg_rotate_en <= 1'b0;
      end

      s_tx_rotate: begin
        tx_bit_count_clear <= 1'b0;
        tx_bit_delay_count_clear <= 1'b1;
        tx_bit_delay_count_inc <= 1'b0;
        tx_data_reg_rotate_en <= 1'b1;
      end

      default: begin
        tx_bit_count_clear <= 1'b0;
        tx_bit_delay_count_clear <= 1'b0;
        tx_bit_delay_count_inc <= 1'b0;
        tx_data_reg_rotate_en <= 1'b0;
      end

    endcase
  end



endmodule




