//-----------------------------------------------------------------------------
//  Mandelbrot Set Generator
//  ECE573 Project, Winter 2008
//  Author:  Eddie Yang
//-----------------------------------------------------------------------------
//------------------------------------------------------------------------------
//  uart_decode.v
//
//  The uart_decode module is the interface between the tty module and the
//    Mandelbrot set generator block.  Its main purpose is to take the ASCII
//    input from the tty module and decode it.  Depending on what that decodes
//    to, a command will be given to the Mandelbrot generator to either pan or
//    zoom in.  When the Mandelbrot generator is finished, a signal is sent back
//    to the tty module which will then echo the input character back to the
//    user which signifies that we are ready to accept another command.
//
//------------------------------------------------------------------------------

`timescale 1 ps / 1 ps

module uart_decode (  input reset_n,
                      input clk,
                      input rx_rdy,                 //data from uart is ready
                      input [7:0] rx_data_reg,      //data from uart
                      input frame_done,             //response from Mand gen

                      output reg echo,
                      output reg start,
                      output reg restart,
                      output reg left,
                      output reg right,
                      output reg up,
                      output reg down,
                      output reg in,
                      output reg out,
                      output reg anim,
                      output reg cursor_en_dis  );

  //----------------------------------------------------------------------------
  //One hot encoding of states.  The idle state is kicked out by rx_rdy.
  //  The decode state decodes the rx_data_reg and also sets which signal
  //  the data_reg corresonds to.
  //
  //The start state checks the start_flag signal and kicks the machine back to
  //  idle if it's not start. Otherwise it proceeds to the wait state.  The wait
  //  waits for the done signal.  At that point, the machine goes into the echo
  //  state who sets the echo signal for the tty machine to echo back to the
  //  terminal.
  //
  //A check for whether the data reg is valid is not needed because the decode
  //  state will only assert a cmd signal if the data reg decodes to a valid
  //  command.  Otherwise, the state machine will continue to the start state
  //  where it will go back to the idle state anyway because an invalid command
  //  is obviously not a start command.
  //----------------------------------------------------------------------------

  //state encoding
  parameter s_idle =        4'b0001;
  parameter s_decode =      4'b0010;
  parameter s_start =       4'b0100;
  parameter s_wait =        4'b1000;

  //parameterized ASCII codes
  parameter asc_start =     8'b01110011;          //s
  parameter asc_restart =   8'b01110010;          //r
  parameter asc_left =      8'h66;                //f
  parameter asc_right =     8'h68;                //h
  parameter asc_up =        8'h74;                //t
  parameter asc_down =      8'h67;                //g
  parameter asc_in =        8'b01101001;          //i
  parameter asc_out =       8'b01101111;          //o
  parameter asc_key_a =     8'h61;                //a
  parameter asc_key_c =     8'h63;                //c

  reg [4:0] current_state;
  reg [4:0] next_state;

  reg start_flag;

  //----------------------------------------------------------------------------
  //implement start_flag
  //----------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      start_flag <= 1'b0;
    else if (clk) begin
      if (start | restart)
        start_flag <= 1'b1;
      else if (frame_done)
        start_flag <= 1'b0;
    end
  end

  //----------------------------------------------------------------------------
  //current state logic
  //----------------------------------------------------------------------------
  always @(negedge reset_n or posedge clk) begin
    if (!reset_n)
      current_state <= s_idle;
    else if (clk)
      current_state <= next_state;
  end

  //----------------------------------------------------------------------------
  //next state logic
  //----------------------------------------------------------------------------
  always @* begin

    case (current_state)

      s_idle : begin
        if (!rx_rdy)
          next_state <= s_idle;
        else
          next_state <= s_decode;
      end

      s_decode : begin
        next_state <= s_start;
      end

      s_start : begin
        if (!start_flag)
          next_state <= s_idle;
        else
          next_state <= s_wait;
      end

      s_wait : begin
        if (!frame_done)
          next_state <= s_wait;
        else
          next_state <= s_idle;
      end

      default : begin
        next_state <= s_idle;
      end

    endcase //case (current_state)
  end //next state logic

  //----------------------------------------------------------------------------
  //output logic
  //----------------------------------------------------------------------------
  always @*  begin

    case (current_state)

      s_idle :  begin
        echo <= 1'b0;
        start <= 1'b0;
        restart <= 1'b0;
        left <= 1'b0;
        right <= 1'b0;
        up <= 1'b0;
        down <= 1'b0;
        in <= 1'b0;
        out <= 1'b0;
        anim <= 1'b0;
        cursor_en_dis <= 1'b0;
      end

      s_decode :  begin
        case (rx_data_reg)

          asc_start: begin
            echo <= 1'b0;
            start <= 1'b1;
            restart <= 1'b0;
            left <= 1'b0;
            right <= 1'b0;
            up <= 1'b0;
            down <= 1'b0;
            in <= 1'b0;
            out <= 1'b0;
            anim <= 1'b0;
            cursor_en_dis <= 1'b0;
          end

          asc_restart: begin
            echo <= 1'b0;
            start <= 1'b0;
            restart <= 1'b1;
            left <= 1'b0;
            right <= 1'b0;
            up <= 1'b0;
            down <= 1'b0;
            in <= 1'b0;
            out <= 1'b0;
            anim <= 1'b0;
            cursor_en_dis <= 1'b0;
          end

          asc_left: begin
            echo <= 1'b1;
            start <= 1'b0;
            restart <= 1'b0;
            left <= 1'b1;
            right <= 1'b0;
            up <= 1'b0;
            down <= 1'b0;
            in <= 1'b0;
            out <= 1'b0;
            anim <= 1'b0;
            cursor_en_dis <= 1'b0;
          end

          asc_right:  begin
            echo <= 1'b1;
            start <= 1'b0;
            restart <= 1'b0;
            left <= 1'b0;
            right <= 1'b1;
            up <= 1'b0;
            down <= 1'b0;
            in <= 1'b0;
            out <= 1'b0;
            anim <= 1'b0;
            cursor_en_dis <= 1'b0;
          end

          asc_up: begin
            echo <= 1'b1;
            start <= 1'b0;
            restart <= 1'b0;
            left <= 1'b0;
            right <= 1'b0;
            up <= 1'b1;
            down <= 1'b0;
            in <= 1'b0;
            out <= 1'b0;
            anim <= 1'b0;
            cursor_en_dis <= 1'b0;
          end

          asc_down: begin
            echo <= 1'b1;
            start <= 1'b0;
            restart <= 1'b0;
            left <= 1'b0;
            right <= 1'b0;
            up <= 1'b0;
            down <= 1'b1;
            in <= 1'b0;
            out <= 1'b0;
            anim <= 1'b0;
            cursor_en_dis <= 1'b0;
          end

          asc_in: begin
            echo <= 1'b1;
            start <= 1'b0;
            restart <= 1'b0;
            left <= 1'b0;
            right <= 1'b0;
            up <= 1'b0;
            down <= 1'b0;
            in <= 1'b1;
            out <= 1'b0;
            anim <= 1'b0;
            cursor_en_dis <= 1'b0;
          end

          asc_out: begin
            echo <= 1'b1;
            start <= 1'b0;
            restart <= 1'b0;
            left <= 1'b0;
            right <= 1'b0;
            up <= 1'b0;
            down <= 1'b0;
            in <= 1'b0;
            out <= 1'b1;
            anim <= 1'b0;
            cursor_en_dis <= 1'b0;
          end

          asc_key_a: begin
            echo <= 1'b1;
            start <= 1'b0;
            restart <= 1'b0;
            left <= 1'b0;
            right <= 1'b0;
            up <= 1'b0;
            down <= 1'b0;
            in <= 1'b0;
            out <= 1'b0;
            anim <= 1'b1;
            cursor_en_dis <= 1'b0;
          end

          asc_key_c: begin
            echo <= 1'b1;
            start <= 1'b0;
            restart <= 1'b0;
            left <= 1'b0;
            right <= 1'b0;
            up <= 1'b0;
            down <= 1'b0;
            in <= 1'b0;
            out <= 1'b0;
            anim <= 1'b0;
            cursor_en_dis <= 1'b1;
          end

          default: begin
            echo <= 1'b1;
            start <= 1'b0;
            restart <= 1'b0;
            left <= 1'b0;
            right <= 1'b0;
            up <= 1'b0;
            down <= 1'b0;
            in <= 1'b0;
            out <= 1'b0;
            anim <= 1'b0;
            cursor_en_dis <= 1'b0;
          end

        endcase //case (rx_data_reg)
      end //decode output logic

      s_start :  begin
        echo <= 1'b0;
        start <= 1'b0;
        restart <= 1'b0;
        left <= 1'b0;
        right <= 1'b0;
        up <= 1'b0;
        down <= 1'b0;
        in <= 1'b0;
        out <= 1'b0;
        anim <= 1'b0;
        cursor_en_dis <= 1'b0;
      end

      s_wait : begin
        if (frame_done)
          echo <= 1'b1;
        else
          echo <= 1'b0;
        start <= 1'b0;
        restart <= 1'b0;
        left <= 1'b0;
        right <= 1'b0;
        up <= 1'b0;
        down <= 1'b0;
        in <= 1'b0;
        out <= 1'b0;
        anim <= 1'b0;
        cursor_en_dis <= 1'b0;
      end

      default: begin
        echo <= 1'b0;
        start <= 1'b0;
        restart <= 1'b0;
        left <= 1'b0;
        right <= 1'b0;
        up <= 1'b0;
        down <= 1'b0;
        in <= 1'b0;
        out <= 1'b0;
        anim <= 1'b0;
        cursor_en_dis <= 1'b0;
      end

    endcase //case (current_state)
  end //output logic

endmodule //uart_decode

