//-----------------------------------------------------------------------------
//  Mandelbrot Set Generator
//  ECE573 Project, Winter 2008
//  Author:  Jesse Armagost
//-----------------------------------------------------------------------------


`timescale 1 ps / 1 ps

`include "vga_defs_1024_768.v"


module vga_controller ( input                             reset_n,
                        input                             vgaclk,
                        input       [3:0]                 r_in,
                        input       [3:0]                 g_in,
                        input       [3:0]                 b_in,
                        output      [`LB2_PIXELS_H - 1:0] h_coord,
                        output      [`LB2_PIXELS_V - 1:0] v_coord,
                        output reg  [3:0]                 r_out,
                        output reg  [3:0]                 g_out,
                        output reg  [3:0]                 b_out,
                        output reg                        h_sync,
                        output reg                        v_sync,
                        output                            sram_wr_start_en  );


  //wires and regs
  reg [(`LB2_PIXELS_H - 1):0] h_count;
  reg [(`LB2_PIXELS_V - 1):0] v_count;

  reg h_count_done;
  reg v_count_done;
  wire h_line_done;
  reg [3:0] h_state;
  reg [3:0] v_state;

  wire h_sync_int;
  wire v_sync_int;
  wire [3:0] r_out_int;
  wire [3:0] g_out_int;
  wire [3:0] b_out_int;

  parameter s_h_sync =          4'b0001;
  parameter s_h_left_blank =    4'b0010;
  parameter s_h_pixels =        4'b0100;
  parameter s_h_right_blank =   4'b1000;

  parameter s_v_sync =          4'b0001;
  parameter s_v_top_blank =     4'b0010;
  parameter s_v_pixels =        4'b0100;
  parameter s_v_bottom_blank =  4'b1000;

  //drive h_sync, v_sync, left_blank and right_blank
  assign h_sync_int = (h_state == s_h_sync) ? 1'b1 : 1'b0;
  assign v_sync_int = (v_state == s_v_sync) ? 1'b1 : 1'b0;

  //drive rgb
  assign r_out_int = (    h_state == s_h_pixels
                      &&  v_state == s_v_pixels ) ? r_in : 4'b0000;

  assign g_out_int = (    h_state == s_h_pixels
                      &&  v_state == s_v_pixels ) ? g_in : 4'b0000;

  assign b_out_int = (    h_state == s_h_pixels
                      &&  v_state == s_v_pixels ) ? b_in : 4'b0000;

  //generate h and v coordinates
  assign h_coord = h_count;
  assign v_coord = v_count;

  //register everything that needs to be registered on its way out
  always @(negedge reset_n or posedge vgaclk) begin
    if (~reset_n) begin
      h_sync <= 0;
      v_sync <= 0;
      r_out <= 0;
      g_out <= 0;
      b_out <= 0;
    end
    else if (vgaclk) begin
      h_sync <= h_sync_int;
      v_sync <= v_sync_int;
      r_out <= r_out_int;
      g_out <= g_out_int;
      b_out <= b_out_int;
    end
  end

  //update horizontal state.  if the counter is done then move to the next
  //  state.  h_count_done has specific meaning depending on the current state.
  always @(negedge reset_n or posedge vgaclk) begin
    if (!reset_n)
      h_state <= s_h_sync;
    else if (vgaclk) begin
      if (h_count_done) begin

        case (h_state)

          s_h_sync:
            h_state <= s_h_left_blank;

          s_h_left_blank:
            h_state <= s_h_pixels;

          s_h_pixels:
            h_state <= s_h_right_blank;

          default:
            h_state <= s_h_sync;

        endcase

      end
    end
  end

  //generate a signal to tell us when a complete horizontal line is done.  we
  //  use this to gate the update of the vertical state.
  assign h_line_done =
                  ((h_state == s_h_right_blank) && h_count_done) ? 1'b1 : 1'b0;


  //update vertical state.  gated by h_line_done because the vertical state
  //should only change when a complete horizontal line is done.
  always @(negedge reset_n or posedge vgaclk) begin
    if (!reset_n)
      v_state <= s_v_sync;
    else if (vgaclk) begin
      if (h_line_done) begin
        if (v_count_done) begin

          case (v_state)

            s_v_sync:
              v_state <= s_v_top_blank;

            s_v_top_blank:
              v_state <= s_v_pixels;

            s_v_pixels:
              v_state <= s_v_bottom_blank;

            default:
              v_state <= s_v_sync;

          endcase

        end
      end
    end
  end


  //horizontal counter
  always @(negedge reset_n or posedge vgaclk) begin
    if (!reset_n)
      h_count <= 0;
    else if (vgaclk) begin
      if (h_count_done)
        h_count <= 0;
      else
        h_count <= h_count + 1'b1;
    end
  end

  //generate the h_count_done signal depending on which horizontal state we are
  //  currently in.
  always @(h_state or h_count) begin

    case (h_state)

      s_h_sync: begin
        if (h_count == `SYNC_WIDTH_H - 1)
          h_count_done <= 1'b1;
        else
          h_count_done <= 1'b0;
      end

      s_h_left_blank: begin
        if (h_count == `LEFT_BLANK - 1)
          h_count_done <= 1'b1;
        else
          h_count_done <= 1'b0;
      end

      s_h_pixels: begin
        if (h_count == `PIXELS_H - 1)
          h_count_done <= 1'b1;
        else
          h_count_done <= 1'b0;
      end

      s_h_right_blank: begin
        if (h_count == `RIGHT_BLANK - 1)
          h_count_done <= 1'b1;
        else
          h_count_done <= 1'b0;
      end

      default:
        h_count_done <= 1'b0;

    endcase

  end


  //vertical counter.  only animate counter when horizontal right sync count is
  //  done.  (once per horizontal line).  h_line_done tells us this condition.
  always @(negedge reset_n or posedge vgaclk) begin
    if (!reset_n)
      v_count <= 0;
    else if (vgaclk) begin
      if (h_line_done) begin
        if (v_count_done)
          v_count <= 0;
        else
          v_count <= v_count + 1'b1;
      end
    end
  end


  //generate the v_count_done signal depending on which vertical state we are
  //  currently in.
  always @(v_state or v_count) begin

    case (v_state)

      s_v_sync: begin
        if (v_count == `SYNC_WIDTH_V - 1)
          v_count_done <= 1'b1;
        else
          v_count_done <= 1'b0;
      end

      s_v_top_blank: begin
        if (v_count == `TOP_BLANK - 1)
          v_count_done <= 1'b1;
        else
          v_count_done <= 1'b0;
      end

      s_v_pixels: begin
        if (v_count == `PIXELS_V - 1)
          v_count_done <= 1'b1;
        else
          v_count_done <= 1'b0;
      end

      s_v_bottom_blank: begin
        if (v_count == `BOTTOM_BLANK - 1)
          v_count_done <= 1'b1;
        else
          v_count_done <= 1'b0;
      end

      default:
        v_count_done <= 1'b0;

    endcase

  end

  //generate the mandelbrot sram write start enable signal.
  assign sram_wr_start_en = (    h_state == s_h_sync
                              || h_state == s_h_right_blank
                              || v_state == s_v_sync
                              || v_state == s_v_top_blank
                              || v_state == s_v_bottom_blank  ) ? 1'b1 : 1'b0;




endmodule


