//-----------------------------------------------------------------------------
//  Mandelbrot Set Generator
//  ECE573 Project, Winter 2008
//  Author:  Jesse Armagost
//-----------------------------------------------------------------------------


`timescale 1 ps / 1 ps

module vga_palette (  input               vga_reset_n,
                      input               vgaclk,
                      input               color_cycle_en,
                      input               vga_vsync,
                      input       [1:0]   palette_select,
                      input       [3:0]   index,
                      output reg  [11:0]  rgb_data        );

  //regs and wires
  reg [5:0] index_1q;
  reg [22:0] lfsr;
  wire animation_tick;
  reg [3:0] animate_offset;
  wire offset_inc_en;
  reg need_offset_inc;
  reg color_cycle_en_1q;
  reg color_cycle_en_2q;
  reg color_cycle_en_3q;
  wire color_cycle_en_now;
  reg color_cycle_en_int;
  reg vsync_edge_det_1q;
  reg vsync_edge_det_2q;
  wire vsync_now;

  //animate the 23 bit lfsr, which gives us a ~0.25s time period.
  always @(negedge vga_reset_n or posedge vgaclk) begin
    if (~vga_reset_n)
      lfsr <= 23'h7fffff;
    else if (vgaclk)
      lfsr <= {lfsr[21:0], lfsr[22] ^ lfsr[17]};
  end

  assign animation_tick = (lfsr == 23'h7fffff) ? 1'b1 : 1'b0;

  //detect the rising edge of vga vsync.  this is when we will increment the
  //  animation offset counter if we need to.
  always @(negedge vga_reset_n or posedge vgaclk) begin
    if (~vga_reset_n) begin
      vsync_edge_det_1q <= 1'b0;
      vsync_edge_det_2q <= 1'b0;
    end
    else if (vgaclk) begin
      vsync_edge_det_1q <= vga_vsync;
      vsync_edge_det_2q <= vsync_edge_det_1q;
    end
  end

  assign vsync_now = (vsync_edge_det_1q & ~vsync_edge_det_2q);

  //remember if we had an animation tick since the last vsync.  if so, we need
  //  to increment the animation offset counter.
  always @(negedge vga_reset_n or posedge vgaclk) begin
    if (~vga_reset_n)
      need_offset_inc <= 1'b0;
    else if (vgaclk) begin
      if (offset_inc_en)
        need_offset_inc <= 1'b0;
      else if (animation_tick)
        need_offset_inc <= 1'b1;
    end
  end

  assign offset_inc_en = (  vsync_now
                          & need_offset_inc
                          & color_cycle_en_int  );

  //maintain the animation offset value.  this value is stepped up whenever we
  //  get an animation tick.  if animation is enabled, we add this value when
  //  we perform a lookup into the palette.  to avoid tearing, we only
  //  increment this counter during vga vertical sync.
  always @(negedge vga_reset_n or posedge vgaclk) begin
    if (~vga_reset_n)
      animate_offset <= 0;
    else if (vgaclk) begin
      if (offset_inc_en)
        animate_offset <= animate_offset + 1'b1;
    end
  end

  //synchronize the color_cycle_en signal, which comes from the sys clock domain.
  //  also add in a rising edge detect, because this synchronized signal may
  //  end up being stretched.
  always @(negedge vga_reset_n or posedge vgaclk) begin
    if (~vga_reset_n) begin
      color_cycle_en_1q <= 0;
      color_cycle_en_2q <= 0;
      color_cycle_en_3q <= 0;
    end
    else if (vgaclk) begin
      color_cycle_en_1q <= color_cycle_en;
      color_cycle_en_2q <= color_cycle_en_1q;
      color_cycle_en_3q <= color_cycle_en_2q;
    end
  end

  assign color_cycle_en_now = (color_cycle_en_2q & ~color_cycle_en_3q);

  //remember whether color cycle animation is enabled or not.  this toggles on
  //  and off every time the 'a' key is pressed and the color_cycle_en signal
  //  toggles.
  always @(negedge vga_reset_n or posedge vgaclk) begin
    if (~vga_reset_n)
      color_cycle_en_int <= 1'b0;
    else if (vgaclk) begin
      if (color_cycle_en_now)
        color_cycle_en_int <= ~color_cycle_en_int;
    end
  end

  //register the incoming palette select and index.  if color cycling is
  //  enabled, we add the animate_offset.
  always @(negedge vga_reset_n or posedge vgaclk) begin
    if (~vga_reset_n)
      index_1q <= 0;
    else if (vgaclk) begin
      if (color_cycle_en_int)
        index_1q <= {palette_select, index + animate_offset};
      else
        index_1q <= {palette_select, index};
    end
  end

  //produce the registered rgb data out
  always @(negedge vga_reset_n or posedge vgaclk) begin
    if (~vga_reset_n)
      rgb_data <= 0;
    else if (vgaclk) begin
      case (index_1q)

        //first palette.  black->blue->purple->white
        default : rgb_data <= {4'h0, 4'h0, 4'h0};
        6'h01   : rgb_data <= {4'h1, 4'h0, 4'h3};
        6'h02   : rgb_data <= {4'h2, 4'h0, 4'h6};
        6'h03   : rgb_data <= {4'h3, 4'h0, 4'h7};
        6'h04   : rgb_data <= {4'h4, 4'h0, 4'h8};
        6'h05   : rgb_data <= {4'h5, 4'h1, 4'hb};
        6'h06   : rgb_data <= {4'h6, 4'h1, 4'he};
        6'h07   : rgb_data <= {4'h6, 4'h2, 4'hf};
        6'h08   : rgb_data <= {4'h7, 4'h2, 4'hf};
        6'h09   : rgb_data <= {4'h8, 4'h6, 4'hf};
        6'h0a   : rgb_data <= {4'ha, 4'ha, 4'hf};
        6'h0b   : rgb_data <= {4'hb, 4'hb, 4'hf};
        6'h0c   : rgb_data <= {4'hc, 4'he, 4'hf};
        6'h0d   : rgb_data <= {4'hd, 4'he, 4'hf};
        6'h0e   : rgb_data <= {4'he, 4'he, 4'hf};
        6'h0f   : rgb_data <= {4'hf, 4'hf, 4'hf};

        //second palette.  black->red->pink->white
        6'h10   : rgb_data <= {4'h0, 4'h0, 4'h0};
        6'h11   : rgb_data <= {4'h2, 4'h0, 4'h0};
        6'h12   : rgb_data <= {4'h4, 4'h0, 4'h0};
        6'h13   : rgb_data <= {4'h5, 4'h0, 4'h1};
        6'h14   : rgb_data <= {4'h7, 4'h0, 4'h2};
        6'h15   : rgb_data <= {4'h9, 4'h0, 4'h2};
        6'h16   : rgb_data <= {4'ha, 4'h0, 4'h2};
        6'h17   : rgb_data <= {4'hb, 4'h0, 4'h3};
        6'h18   : rgb_data <= {4'hd, 4'h1, 4'h3};
        6'h19   : rgb_data <= {4'hd, 4'h2, 4'h5};
        6'h1a   : rgb_data <= {4'hd, 4'h3, 4'h7};
        6'h1b   : rgb_data <= {4'hd, 4'h6, 4'ha};
        6'h1c   : rgb_data <= {4'hd, 4'h7, 4'hb};
        6'h1d   : rgb_data <= {4'hd, 4'h9, 4'hc};
        6'h1e   : rgb_data <= {4'hd, 4'hc, 4'hd};
        6'h1f   : rgb_data <= {4'hd, 4'hf, 4'hf};

        //third palette.  black->orange->lt. orange->white
        6'h20   : rgb_data <= {4'h0, 4'h0, 4'h0};
        6'h21   : rgb_data <= {4'h3, 4'h1, 4'h0};
        6'h22   : rgb_data <= {4'h4, 4'h1, 4'h0};
        6'h23   : rgb_data <= {4'h5, 4'h1, 4'h0};
        6'h24   : rgb_data <= {4'h7, 4'h1, 4'h1};
        6'h25   : rgb_data <= {4'h9, 4'h2, 4'h1};
        6'h26   : rgb_data <= {4'ha, 4'h2, 4'h1};
        6'h27   : rgb_data <= {4'hd, 4'h2, 4'h1};
        6'h28   : rgb_data <= {4'hf, 4'h3, 4'h2};
        6'h29   : rgb_data <= {4'hf, 4'h4, 4'h4};
        6'h2a   : rgb_data <= {4'hf, 4'h6, 4'h7};
        6'h2b   : rgb_data <= {4'hf, 4'h8, 4'h9};
        6'h2c   : rgb_data <= {4'hf, 4'ha, 4'ha};
        6'h2d   : rgb_data <= {4'hf, 4'hb, 4'hd};
        6'h2e   : rgb_data <= {4'hf, 4'hc, 4'he};
        6'h2f   : rgb_data <= {4'hf, 4'hf, 4'hf};

        //fourth palette.  ??
        6'h30   : rgb_data <= {4'h0, 4'h0, 4'h0};
        6'h31   : rgb_data <= {4'h3, 4'h2, 4'h4};
        6'h32   : rgb_data <= {4'h4, 4'h3, 4'h8};
        6'h33   : rgb_data <= {4'h5, 4'h4, 4'hc};
        6'h34   : rgb_data <= {4'h7, 4'h3, 4'hd};
        6'h35   : rgb_data <= {4'h9, 4'h2, 4'h9};
        6'h36   : rgb_data <= {4'hc, 4'h1, 4'h5};
        6'h37   : rgb_data <= {4'hf, 4'h2, 4'h1};
        6'h38   : rgb_data <= {4'hd, 4'h3, 4'h0};
        6'h39   : rgb_data <= {4'ha, 4'h4, 4'h4};
        6'h3a   : rgb_data <= {4'h9, 4'h3, 4'h8};
        6'h3b   : rgb_data <= {4'h7, 4'h2, 4'hc};
        6'h3c   : rgb_data <= {4'h5, 4'h1, 4'hd};
        6'h3d   : rgb_data <= {4'h4, 4'h2, 4'h9};
        6'h3e   : rgb_data <= {4'h3, 4'h3, 4'h5};
        6'h3f   : rgb_data <= {4'h2, 4'h2, 4'h1};

      endcase
    end
  end

endmodule
